import math
import pprint
import difflib

import pylibelf

def elfdiff(obj1, obj2, mask=None):
    L1 = prettify(obj1, mask=mask).splitlines()
    L2 = prettify(obj2, mask=mask).splitlines()
    return '\n'.join(difflib.unified_diff(L1, L2, n=10))
    
def _format_collection(obj, indent, mask, oneline, maxlength):
    indent += 1
    sep = '\n'.ljust(indent)
    g = (val for o in obj 
                 if not mask.matchvalue(o) 
                     for val in [prettify(o, indent, mask, oneline, maxlength)]
                         if val)
    result = sep.join(g)
    if not result and obj:
        return '...'
    return result
    
def prettify(obj, indent=1, mask=None, oneline=True, maxlength=80):
    """
    return the pretty output of an object
    `indent` how much to indent the output
    `mask` determines what to keep and what to ditch, 
           see the PrettifyMask class
    `oneline` indicates whether it should be attempted to try to format 
              everything on one line -- if not too long -- or whether 
              to always use a new line for an attribute
    """
    linelength = maxlength - indent
    
    if mask is None:
        mask = pylibelf.settings.mask
    
    if isinstance(obj, (tuple, list)):
        fmt = _format_collection(obj, indent, mask, oneline, maxlength)
        return '[%s]' % fmt
    if isinstance(obj, (int, long)):
        # Don't append an annoying 'L' like hex() does for longs
        return '0x%02x' % obj
    if isinstance(obj, basestring):
        obj = repr(obj)
        endidx = len(obj)
        if linelength < 20:
            endidx = 20
        elif len(obj) > linelength:
            endidx = linelength
    
        if endidx < len(obj):
            obj = '%s ...%s' % (obj[:endidx], obj[0])

        return obj

    attrs = getattr(obj, 'attributes', None)
    if attrs is None:
        return str(obj)
    
    return obj._str(indent, linelength, mask, oneline, maxlength)


def pretty_print(*args, **kwargs):
    for obj in args:
        print(prettify(obj, **kwargs))


class PrettifyMask(object):
    """
    Mask that can be used for not printing certain kind of things in a 
    PrettyPrintBase instance. A mask can be inverted with the ~ operator.
    An inverted mask operates only on attributes that are set, so acts more 
    like a "keep this stuff" object than a mask. Only conditions that are 
    specified are used in the matching.
    """
    
    def __init__(self, attrs=None, values=None, classes=None, attrkeys=None, 
                 valuekeys=None, invert=False):
        """
        `attrs` the attribute names to omit
        `values` the attribute values to omit
        `classes` instances of these classes will be omitted. Must be a tuple
        `attrkeys` callables that determine whether to include an attribute
        `valuekeys` callables that determine whether to include a value
        """
        self.attrs = attrs
        self.values = values
        self.classes = classes
        self.attrkeys = attrkeys
        self.valuekeys = valuekeys
        self.invert = invert
    
    def _match(self, value, maskvalues, callables):
        return ((maskvalues and value in maskvalues) or
                (callables and any(f(value) for f in callables)))
    
    def _return_result(self, result):
        if self.invert:
            return not result
        else:
            return result
    
    def matchattr(self, attr):
        "returns whether attribute name `attr` is masked"
        if self.emptyattrs:
            return False
        result = self._match(attr, self.attrs, self.attrkeys)
        return self._return_result(result)
    
    def matchvalue(self, value):
        "returns whether value `value` is masked"
        if self.emptyvalues:
            return False
        classes = self.classes
        result = (self._match(value, self.values, self.valuekeys) or
                  (classes and isinstance(value, classes)))
        return self._return_result(result)
    
    def match(self, obj, attr):
        "returns whether attribute `attr` (given as a string) is masked"
        matched_attr = self.matchattr(attr) 
        matched_value = self.matchvalue(getattr(obj, attr))
        if self.invert:
            if self.emptyattrs:
                return matched_value
            elif self.emptyvalues:
                return matched_attr
            else:
                return matched_attr and matched_value
        else:
            return matched_attr or matched_value
    
    def __invert__(self):
        return type(self)(**dict(vars(self), invert=not self.invert))
    
    @property
    def emptyattrs(self):
        return all(c is None for c in (self.attrs, self.attrkeys))
    
    @property
    def emptyvalues(self):
        conds = self.values, self.valuekeys, self.classes
        return all(c is None for c in conds)
    
    @property
    def empty(self):
        return self.emptyattrs and self.emptyvalues

_NOERROR = object()
def find(L, errors=_NOERROR, index=False, **kwargs):
    if errors is _NOERROR:
        errors = pylibelf.settings.find_errors

    try:
        t = _findall_idx(L, **kwargs).next()
        if index:
            return t
        return t[1]
    except StopIteration:
        if errors:
            raise IndexError("%s not found" % (kwargs,))
        return None

def findall(L, **kwargs):
    return [t[1] for t in _findall_idx(L, **kwargs)]

def _findall_idx(L, **kwargs):
    if not kwargs:
        raise ValueError('Must provide at least one keyword argument')

    items = kwargs.items()
    for idx, obj in enumerate(L):
        if all([getattr(obj, attr) == value for attr, value in items]):
            yield idx, obj


def getnbits(x):
    "return the number of bits is needed to represent a number"
    if x == 0:
        return 0
        
    return int(math.floor(math.log(x, 2)) + 1)

def getpow(x):
    "return the exponent n of x with base 2 (such that 2 ** n == x)"
    nbits = getnbits(x) - 1
    assert 1 << nbits == x
    return nbits


class Type(object):
    """
    Class for representing constants with their C variable name and 
    documentation
    """
    
    def __init__(self, value, representation=None, varname=None):
        self.value = value
        self.representation = representation
        self.varname = varname
    
    def __eq__(self, other):
        if hasattr(other, 'varname') and None not in (self.varname,
                                                      other.varname):
            return self.varname == other.varname
            
        return other == self.value
    
    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return hash(self.value)
    
    def __repr__(self):
        representation = self.representation
        if representation is None:
            representation = self.value
            
        if self.varname is None:
            return 'Type(%r)' % (representation,)
        else:
           return "Type(%s, %r)" % (self.varname, representation)
    
    def __int__(self):
        return int(self.value)
    

class Flags(int):
    "Pretty repr() class for flags"
    
    def __new__(cls, flags, flagdict):
        return super(Flags, cls).__new__(cls, flags)
        
    def __init__(self, flags, flagdict):
        self.flags = flags
        self.flagdict = flagdict

    def _repr_helper(self):
        if self.flags > 0:
            nbits = getnbits(self.flags)
        else:
            nbits = 0
            
        leftover = self.flags

        for nbit in xrange(nbits, -1, -1):
            flag = self.flags & (1 << nbit)
            if flag in self.flagdict:
                try:
                    representation = self.flagdict[flag].varname
                except AttributeError:
                    representation = self.flagdict[flag]
                    
                yield str(representation)
                leftover &= ~flag
        
        if leftover:
            yield hex(leftover)
    
    def __repr__(self):
        return '|'.join(self._repr_helper()) or '0x0'
